﻿module GameStateGame
open System.Drawing
open System.Drawing.Imaging
open GameState

type game_mode =
    | Mode_Menu
    | Mode_Game

type boat_type =
    | Boat_Red
    | Boat_Green
type boat = {
    boat_type : boat_type;
    boat_x : float32;
    boat_y : float32;
    boat_vx : float32;
    boat_vy : float32;
    boat_spd_mul : float32;
    boat_dir : float32 * float32;
    boat_age : float32
}

type glitter = {
    glitter_x : float32;
    glitter_y : float32;
    glitter_alpha : float32;
    glitter_age : float32;
}

// Creates a new boat structure
let new_boat t x y =
    { boat_type = t;
      boat_x = x;
      boat_y = y;
      boat_vx = 0.f;
      boat_vy = 0.f;
      boat_spd_mul = 1.f;
      boat_dir = Vec2.create 1.f 0.f;
      boat_age = 0.f }

let new_game (scr_w:int) (scr_h:int) =
    let fluid_scale = 8
    let fluid_scale_f = float32 fluid_scale
    let fluid_w = (scr_w / fluid_scale) + 1
    let fluid_h = (scr_h / fluid_scale) + 2
    let fluid = ref (Fluid.create fluid_w fluid_h)
    let fluid_bitmap = new Bitmap(fluid_w, fluid_h, PixelFormat.Format24bppRgb)
    
    let island = Islands.create scr_w scr_h
    
    // The x coordinate for the win area
    let win_area_w = 30.f
    let win_area_x = (float32 scr_w) - win_area_w
    let win_area_div_y = (float32 scr_h) * 0.5f
    
    // Returns the velocity at the given position
    let get_velocity x y =
        Fluid.get_velocity_smooth !fluid (x / fluid_scale_f) (y / fluid_scale_f)
        
    // Tests if the given position is blocked
    let is_blocked_bounded x y =
        let xi = int (x / fluid_scale_f)
        let yi = int (y / fluid_scale_f)
        (Fluid.get_blocked !fluid xi yi)
    let is_blocked x y =
        (x < win_area_x) && (is_blocked_bounded x y)
        
    // Updates the blocked positions in the fluid
    let update_blocked () =
        for fy = 0 to fluid_h - 1 do
            for fx = 0 to fluid_w - 1 do
                let is_full = Islands.is_full island (fx * fluid_scale) (fy * fluid_scale) fluid_scale fluid_scale
                Fluid.set_blocked !fluid fx fy is_full
    update_blocked ()

    // Boat spawn update function
    let boats = ref []
    
    let new_boat_spawn_game () =
        let time_left = ref 3.f
        let spawned_total = ref 0
        let spawns_left = ref 0
        let spawn_y = ref 0.f
        let spawn_type = ref Boat_Red
        fun delta_t ->
            time_left := !time_left - delta_t
            
            // If there are too many boats, halt the spawning
            if List.length !boats > 512 && !time_left <= 0.f then
                time_left := 1.f
                
            // Need to spawn more
            if !time_left <= 0.f then
                // Need a new spawn state
                if !spawns_left <= 0 then
                    let spawn_count = 1 + (int (Util.pow (float !spawned_total) 0.3f))
                    spawns_left := spawn_count
                    spawn_y := Util.rand_float 50.f ((float32 scr_h) - 50.f)
                    spawn_type :=
                        (*match Util.rand_int 0 2 with
                        | 0 -> Boat_Red
                        | _ -> Boat_Green*)
                        match !spawn_type with
                        | Boat_Red -> Boat_Green
                        | Boat_Green -> Boat_Red
                    
                // Spawn a boat
                incr spawned_total
                boats := (new_boat !spawn_type 10.f !spawn_y) :: !boats
                
                // New pause
                decr spawns_left
                time_left := 0.5f
                if !spawns_left <= 0 then
                    time_left := 6.f - (Util.pow (float !spawned_total) 0.35f)
                    if !time_left < 0.5f then
                        time_left := 0.5f

    let new_boat_spawn_menu () =
        fun delta_t -> ()
                        
    let update_boat_spawn = ref (new_boat_spawn_menu ())

    // Draws a single triangle
    let draw_tri =
        let pts = Array.create 3 PointF.Empty
        fun (g:Graphics) brush (x1, y1) (x2, y2) (x3, y3) ->
            pts.[0].X <- x1; pts.[0].Y <- y1
            pts.[1].X <- x2; pts.[1].Y <- y2
            pts.[2].X <- x3; pts.[2].Y <- y3
            g.FillPolygon(brush, pts)
    
    // Draws a single boat    
    let draw_boat g boat =
        let pos = (boat.boat_x, boat.boat_y)
        let dir = boat.boat_dir
        let up = Vec2.get_normal dir
        let behind = Vec2.add_mul pos dir -3.f
        let pt1 = (Vec2.add_mul behind up 3.f)
        let pt2 = (Vec2.add_mul behind up -3.f)
        let pt3 = (Vec2.add_mul pos dir 6.f)
        let s = 3.f
        draw_tri g Brushes.Black (Vec2.scale_away pt1 pos s) (Vec2.scale_away pt2 pos s)(Vec2.scale_away pt3 pos s)
        match boat.boat_type with
        | Boat_Red -> draw_tri g Brushes.Red pt1 pt2 pt3
        | Boat_Green -> draw_tri g Brushes.LightGreen pt1 pt2 pt3
    
    // Creates a new glitter
    let glitter = ref []
    
    let new_glitter () =
        let g = {
            glitter_x = Util.rand_float 0.f (float32 scr_w);
            glitter_y = Util.rand_float 0.f (float32 scr_h);
            glitter_alpha = 0.f;
            glitter_age = 0.f
        }
        glitter := g :: !glitter
    
    // Update the glitter
    let update_glitter delta_t =
        let update gl =
            let vel = Vec2.scale (get_velocity gl.glitter_x gl.glitter_y) 100000.f
            let spd = Vec2.abs vel
            let vel_alpha = Util.map_range 0.f 100.f 0.f 0.85f spd
            let time_alpha = Util.map_range 7.5f 10.f 1.f 0.f gl.glitter_age
            let alpha = vel_alpha * time_alpha
            let vx, vy = vel
            { glitter_x = gl.glitter_x + vx * delta_t;
              glitter_y = gl.glitter_y + vy * delta_t;
              glitter_alpha = Util.lerp gl.glitter_alpha alpha 0.1f;
              glitter_age = gl.glitter_age + delta_t }
        let filter gl =
            (gl.glitter_age < 10.f) && 
            not (is_blocked gl.glitter_x gl.glitter_y) && 
            (gl.glitter_x < (float32 scr_w))
        glitter :=
            !glitter |>
            List.map update |>
            List.filter filter
            
    // Draws the glitter
    let draw_glitter (g:Graphics) =
        let draw gl =
            let br = new SolidBrush(Color.FromArgb(Util.color_f2i gl.glitter_alpha, 200, 200, 255))
            g.FillRectangle(br, gl.glitter_x - 1.f, gl.glitter_y - 1.f, 2.f, 2.f)
        List.iter draw !glitter
    
    // DEBUG: Add some intial blockers   
    (*for y = 0 to 20 do
        for x = 0 to 3 do
            Fluid.set_blocked !fluid (20 + x) (0 + y) true
            Fluid.set_blocked !fluid (30 + x) (fluid_h - 1 - y) true
            Fluid.set_blocked !fluid (50 + x) (0 + y) true
            Fluid.set_blocked !fluid (80 + x) (40 + y) true*)
    
    let mouse_x = ref 0.f
    let mouse_y = ref 0.f
    
    // Goal area alphas
    let goal_red_good = Color.FromArgb(130, 255, 200, 200)
    let goal_red_bad = Color.FromArgb(130, 125, 30, 30)
    let goal_red_normal = Color.FromArgb(100, 255, 100, 100)
    let goal_red_col = ref goal_red_normal
    
    let goal_green_good = Color.FromArgb(130, 200, 255, 200)
    let goal_green_bad = Color.FromArgb(130, 30, 125, 30)
    let goal_green_normal = Color.FromArgb(100, 100, 255, 100)
    let goal_green_col = ref goal_green_normal
    
    // Score and lives
    let score = ref 0
    let hiscore = ref (Storage.get_int "hiscore" 0)
    let lives = ref 0
    let mode = ref Mode_Menu
    let mode_time = ref 0.f
    let mode_changing = ref false
    
    let change_mode new_mode =
        mode := new_mode
        mode_time := 0.f
        mode_changing := false
    
    // Starts a new game
    let new_game () =
        // Reset the baots
        boats := []
        update_boat_spawn := new_boat_spawn_game ()
        
        // Reset the islands
        Islands.clear island
        update_blocked ()
        
        // Reset the score & lives
        score := 0
        lives := 5
        change_mode Mode_Game

    // Starts the menu mode
    let enter_menu () =
        // Remove the boats
        boats := []
        update_boat_spawn := new_boat_spawn_menu ()
        
        // Save the hiscore
        Storage.set_int "hiscore" !hiscore
        
        // Change into the menu
        change_mode Mode_Menu
    
    // Tests if a game is running
    let is_playing () =
        (!mode = Mode_Game) && (!lives > 0)
    
    // Gives a score
    let give_score () =
        if is_playing () then
            incr score
            
            // New hiscore?
            if !score > !hiscore then
                hiscore := !score
            
            // Give additional lives periodically
            if !score % 25 = 0 then
                incr lives
        
    // Loses a life
    let lose_life () =
        if is_playing () then
            decr lives
    
    // Fonts
    let write_gfx (g:Graphics) font (color:Color) str (x:float32) (y:float32) =
        g.DrawString(str, font, new SolidBrush(Color.FromArgb(int color.A, 0, 0, 0)), x + 2.f, y + 2.f)
        g.DrawString(str, font, new SolidBrush(color), x, y)
    
    let write_center (g:Graphics) font (color:Color) str (y:float32) =
        let s = g.MeasureString(str, font)
        let x = ((float32 scr_w) - s.Width) * 0.5f
        write_gfx g font color str x y
    
    // UI elements
    let font_score = new Font("Microsoft Sans Serif", 16.f)
    let font_menu_title = new Font("Microsoft Sans Serif", 50.f)
    let font_menu_credits = new Font("Microsoft Sans Serif", 8.f)
    let font_menu_text = new Font("Microsoft Sans Serif", 14.f)
    
    let score_alpha = ref 0.f
    let menu_alpha = ref 1.f
    
    // The game state draw function
    let draw (g:Graphics) =
        g.Clear Color.Black
        
        // Generate & draw the fluid field
        for y = 0 to fluid_h - 1 do
            for x = 0 to fluid_w - 1 do
                //let dens = Fluid.get_density !fluid x y
                let velx, vely = Fluid.get_velocity !fluid x y
                let dens = sqrt (velx * velx + vely * vely)
                let v = Util.color_f2i (0.3f + dens * 400.f)
                fluid_bitmap.SetPixel(x, y, Color.FromArgb(v * 180 / 256, v * 200 / 256, v))
                // DEBUG: Show the blocked area
                (*if Fluid.get_blocked !fluid x y then
                    fluid_bitmap.SetPixel(x, y, Color.FromArgb(255, 255, 255))*)
                    
        g.DrawImage(fluid_bitmap, 0, 0, fluid_bitmap.Width * fluid_scale, fluid_bitmap.Height * fluid_scale)

        // Draw the islands
        g.DrawImage(Islands.get_bitmap island, 0, 0)

        // Draw the glibber
        draw_glitter g

        // Draw the boats
        List.iter (draw_boat g) !boats

        // Draw the goal areas
        let goal_area_x = win_area_x + Util.map_range 0.f 1.f 0.f (win_area_w + 40.f) !menu_alpha
        g.FillRectangle(new SolidBrush(!goal_red_col), goal_area_x, 0.f, win_area_w, win_area_div_y)
        g.FillRectangle(new SolidBrush(!goal_green_col), goal_area_x, win_area_div_y, win_area_w, win_area_div_y)

        // Show the score
        let score_color = Color.FromArgb(Util.color_f2i !score_alpha, 255, 255, 255)
        write_gfx g font_score score_color (sprintf "Boats: %d" !lives) 10.f 10.f
        write_gfx g font_score score_color (sprintf "Score: %d" !score) 10.f (10.f + (float32 font_score.Height))
        write_gfx g font_score score_color (sprintf "Hiscore: %d" !hiscore) 10.f (10.f + (float32 font_score.Height) * 2.f)

        // Show the menu texts
        if !menu_alpha > 0.01f then
            // Darken the background
            let darken_alpha = Util.map_range 0.f 1.f 0.f 0.6f !menu_alpha
            let darken_col = Color.FromArgb(Util.color_f2i darken_alpha, 0, 0, 0)
            g.FillRectangle(new SolidBrush(darken_col), 0, 0, scr_w + 50, scr_h + 50)
            
            // The title text
            let col_title = Color.FromArgb(Util.color_f2i !menu_alpha, 230, 230, 255)
            write_center g font_menu_title col_title "Dare Islands" 40.f
            
            // The credits text
            let col_credits = Color.FromArgb(Util.color_f2i !menu_alpha, 220, 210, 255)
            write_center g font_menu_credits col_credits "Ludum Dare 48h #17 competition entry" 110.f
            write_center g font_menu_credits col_credits "25.4.2010 -- Jere \"XMunkki\" Sanisalo -- http://www.xmunkki.org/" (110.f + (float32 font_menu_credits.Height))
            
            // Help text
            let col_help = Color.FromArgb(Util.color_f2i !menu_alpha, 220, 250, 225)
            let txt_y = ref 200.f
            let write str =
                if str <> "" then
                    write_center g font_menu_text col_help str !txt_y
                txt_y := !txt_y + (float32 font_menu_text.Height)
            write "Use the left mouse button to paint islands and"
            write "the right mouse button to remove islands."
            write ""
            write "Steer the red boats into the red zone and"
            write "the green boats into the green zone."
            write "Boats must not hit the ground or wrong goal zone."
            write ""
            write ""
            write ""
            write "Press the left mouse button to start!"

        // Show the velocity under the mouse
        (*let vx, vy = get_velocity !mouse_x !mouse_y
        g.DrawLine((new Pen(Color.Red, 2.f)), !mouse_x, !mouse_y, (!mouse_x + vx * 50000.f), (!mouse_y + vy * 50000.f))*)

    // The game state update function
    let update delta_t (mouse:mouse_state) =
        mouse_x := mouse.mouse_x
        mouse_y := mouse.mouse_y
        
        // Update the goal area colors
        goal_red_col := Util.lerp_color !goal_red_col goal_red_normal 0.1f
        goal_green_col := Util.lerp_color !goal_green_col goal_green_normal 0.1f
        
        // Update the fluid
        fluid := Fluid.update !fluid
        
        // Update the glitter
        new_glitter ()
        update_glitter delta_t        
        
        // Update the boats
        let get_speed x y =
            let vx, vy = get_velocity x y
            //let mag = sqrt (vx * vx + vy * vy)
            //let scale = 100.f
            //if mag < 0.00001f then (scale, 0.f) else (vx * scale / mag, vy * scale / mag)

            let scale = 10000.f
            let vx = vx * scale 
            let vy = vy * scale
            let spd = sqrt (vx * vx + vy * vy)
            let min_spd = 30.f
            if spd < 0.001f then
                (min_spd, 0.f)
            elif spd < min_spd then
                (vx * min_spd / spd, vy * min_spd / spd)
            else
                (vx, vy)
        
        let get_wall_avoidance x y =
            let steps = 8
            let step = (float32 System.Math.PI) * 2.f / (float32 steps)
            let test_dist = 25.f
            let mutable ang = 0.f
            let mutable nx = 0.f
            let mutable ny = 0.f
            for i = 1 to steps do
                let dx = cos ang
                let dy = sin ang
                if is_blocked (x + dx * test_dist) (y + dy * test_dist) then
                    nx <- nx - dx
                    ny <- ny - dy
                ang <- ang + step
            Vec2.unit_or_zero (nx, ny)
        
        let update_boat boat =
            let vel_mul = delta_t
            
            // Get the new speed
            let vx, vy = get_speed boat.boat_x boat.boat_y
            let vx, vy = Vec2.add_mul (vx, vy) (get_wall_avoidance boat.boat_x boat.boat_y) 30.f
            let vx = Util.lerp boat.boat_vx vx 0.2f
            let vy = Util.lerp boat.boat_vy vy 0.2f
            
            // Get the new boat dir
            let new_dir = Vec2.unit_or_zero (vx, vy)
            let new_dir = Vec2.unit_or_zero (Vec2.lerp boat.boat_dir new_dir 0.2f)
            
            // Get the speed multiplier
            let speed_mul =
                let test_dist d =
                    let tx, ty = Vec2.add_mul (boat.boat_x, boat.boat_y) boat.boat_dir d
                    is_blocked tx ty
                if test_dist 10.f || test_dist 20.f || test_dist 30.f then 
                    0.f
                else
                    1.f
            let new_speed_mul = Util.lerp boat.boat_spd_mul speed_mul 0.07f
            
            let vx = vx * new_speed_mul
            let vy = vy * new_speed_mul
            
            // Get the new position
            let nx = boat.boat_x + vx * delta_t
            let ny = boat.boat_y + vy * delta_t
            { boat_type = boat.boat_type;
              boat_x = nx;
              boat_y = ny;
              boat_vx = vx;
              boat_vy = vy;
              boat_spd_mul = new_speed_mul;
              boat_dir = new_dir;
              boat_age = boat.boat_age + delta_t }
        boats := List.map update_boat !boats
        
        // Remove the boats that are done
        let test_boat_remove boat =
            if boat.boat_x >= win_area_x then
                // Goal conditions
                match boat.boat_type with
                | Boat_Red when boat.boat_y <= win_area_div_y ->
                    goal_red_col := goal_red_good 
                    give_score ()
                | Boat_Red ->
                    goal_green_col := goal_green_bad
                    lose_life ()
                | Boat_Green when boat.boat_y >= win_area_div_y ->
                    goal_green_col := goal_green_good 
                    give_score ()
                | Boat_Green ->
                    goal_red_col := goal_red_bad
                    lose_life ()
                false
            elif is_blocked boat.boat_x boat.boat_y then
                // Boat hit something
                lose_life ()
                false
            else
                true
        boats := List.filter test_boat_remove !boats
        
        // Create new boats?
        (*if mouse.mouse_right_pressed then
            boats := (new_boat Boat_Red mouse.mouse_x mouse.mouse_y) :: !boats*)
        
        !update_boat_spawn delta_t
        
        // Island painting
        if is_playing () then
            let do_painting dir =
                Islands.random_add island (int mouse.mouse_x) (int mouse.mouse_y) 6500 200.f (0.05f * dir)
                Islands.random_add island (int mouse.mouse_x) (int mouse.mouse_y) 250 200.f (-0.05f * dir)
                update_blocked ()
            
            let painting_dir =
                if mouse.mouse_left_down then
                    1.f
                elif mouse.mouse_right_down then
                    -1.f
                else
                    0.f
            if painting_dir <> 0.f then
                do_painting painting_dir
        
        // Mode specific update
        mode_time := !mode_time + delta_t
        match !mode with
        | Mode_Menu ->
            score_alpha := Util.lerp !score_alpha 0.25f 0.05f
            menu_alpha := Util.lerp !menu_alpha 1.f 0.1f
            
            // Start a new game?
            if mouse.mouse_left_pressed && !mode_time > 0.2f then
                mode_changing := true
            if mouse.mouse_left_released && !mode_changing then
                new_game ()
        | Mode_Game ->
            score_alpha := Util.lerp !score_alpha 1.f 0.1f
            menu_alpha := Util.lerp !menu_alpha 0.f 0.075f
            
            // End of game?
            if !lives = 0 then
                enter_menu ()
        
        // Done        
        GSU_Normal
    
    // Return the game state
    { gs_draw = draw; gs_update = update }